#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>

#include <config.h>

static void lo_die(lua_State *L, char const *e)
{
    lua_pushstring(L, e);
    lua_error(L);
}

int lua_pledge(lua_State *L)
{
    char const *farg = NULL;
    char const *earg = NULL;
    int ret = 0;
    int top = 0;

#ifndef HAVE_PLEDGE
    lo_die(L, "pledge: not supported");
#endif

    top = lua_gettop(L);

    luaL_argcheck(L, lua_isstring(L, 1), 1,
                  "pledge: first argument must be string");
    if (top == 2) {
        luaL_argcheck(L, lua_isstring(L, 2), 1,
                      "pledge: second argument must be string");
    } else if (top > 2) {
        lo_die(L, "pledge: too many arguments");
    }

    farg = lua_tostring(L, 1);
    earg = lua_tostring(L, 2);

    ret = pledge(farg, earg);

    lua_pushnumber(L, ret);
    if (ret == -1) {
        lua_pushstring(L, strerror(errno));
        return 2;
    }

    return 1;
}

int lua_arc4random(lua_State *L)
{
#ifndef HAVE_ARC4RANDOM
    lo_die(L, "arc4random: not supported");
#endif

    lua_pushnumber(L, arc4random());
    return 1;
}

int lua_arc4random_uniform(lua_State *L)
{
    uint32_t up = 0;

#ifndef HAVE_ARC4RANDOM_UNIFORM
    lo_die(L, "arc4random_uniform: not supported");
#endif

    luaL_argcheck(L, lua_isnumber(L, 1), 1,
                  "pledge: first argument must be string");

    up = lua_tonumber(L, 1);
    lua_pushnumber(L, arc4random_uniform(up));

    return 1;
}

#if !defined LUA_VERSION_NUM || LUA_VERSION_NUM==501
/* Adapted from Lua 5.2.0
 */
static void luaL_setfuncs (lua_State *L, const luaL_Reg *l, int nup) {
    luaL_checkstack(L, nup+1, "too many upvalues");
    for (; l->name != NULL; l++) {  /* fill the table with given functions */
        int i;
        lua_pushstring(L, l->name);
        for (i = 0; i < nup; i++) {  /* copy upvalues to the top */
            lua_pushvalue(L, -(nup+1));
        }
        lua_pushcclosure(L, l->func, nup);  /* closure with those upvalues */
        lua_settable(L, -(nup + 3));
    }
    lua_pop(L, nup);  /* remove upvalues */
}
#endif

static const struct luaL_Reg l_pledge[] = {
    { "pledge", lua_pledge },
    { "arc4random", lua_arc4random },
    { "arc4random_uniform", lua_arc4random_uniform },
    { NULL, NULL },
};

int luaopen_openbsd(lua_State *L)
{
    lua_newtable(L);
    luaL_setfuncs(L, l_pledge, 0);

    return 1;
}
